# If the sabayon user don't exist, add them:

if ! grep -q sabayon /etc/passwd; then
        /usr/sbin/useradd -c "Sabayon user" -d /var/sabayon -g nobody -s /sbin/false sabayon
fi    




if [ -x /usr/bin/update-mime-database ]; then
  /usr/bin/update-mime-database /usr/share/mime &> /dev/null
fi

if [ -x /usr/bin/update-desktop-database ]; then
  /usr/bin/update-desktop-database 1> /dev/null 2> /dev/null
fi

if [ -x /usr/bin/gtk-update-icon-cache ]; then
  /usr/bin/gtk-update-icon-cache  -f -q /usr/share/icons/hicolor 1> /dev/null 2> /dev/null
fi

